#! -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2020 IndiScale GmbH <info@indiscale.com>
# Copyright (C) 2020 Timm Fitschen <f.fitschen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""unauthenticated.

An Authenticator which suppresses any authentication and also ignores auth_token
cookies.
"""
from __future__ import absolute_import, print_function, unicode_literals

from ...exceptions import LoginFailedError
from .interface import AbstractAuthenticator, CaosDBServerConnection


def get_authentication_provider():
    """get_authentication_provider.

    Return an authenticator which only uses a pre-supplied authentication
    token.

    Returns
    -------
    AuthTokenAuthenticator
    """
    return Unauthenticated()


class Unauthenticated(AbstractAuthenticator):
    """Unauthenticated.

    Subclass of AbstractAuthenticator which suppresses any authentication and
    also ignores auth_token cookies.

    Methods
    -------
    login
    logout
    configure
    on_request
    on_response
    """

    def __init__(self):
        super(Unauthenticated, self).__init__()
        self.auth_token = None
        self._connection = None

    def login(self):
        self._login()

    def _login(self):
        raise LoginFailedError("This LinkAhead client is configured to stay "
                               "unauthenticated. Change your "
                               "`password_method` and provide an "
                               "`auth_token` or credentials if you want "
                               "to authenticate this client.")

    def logout(self):
        self._logout()

    def _logout(self):
        self.auth_token = None

    def configure(self, **config):
        self.auth_token = None

    def on_request(self, method, path, headers, **kwargs):
        # pylint: disable=unused-argument
        """on_request.

        This implementation does not attempt to login or authenticate in any
        form.

        Parameters
        ----------
        method
            unused
        path
            unused
        headers
            unused
        **kwargs
            unused
        """
        pass

    def on_response(self, response):
        # pylint: disable=unused-argument
        """on_response.

        This implementation ignores any auth_token cookie sent by the server.

        Parameters
        ----------
        response
            unused
        """
        pass



# -*- encoding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""Classes and functions to immitate a LinkAhead server without actually setting
up a tcp connection."""

from __future__ import unicode_literals, print_function
from io import StringIO
from .connection import CaosDBServerConnection, CaosDBHTTPResponse


def _request_log_message(method, path, headers, body, **kwargs):
    return "{method}: {path}{nl0}{kwargs}{nl1}{headers}{nl2}{body}".format(
        method=method,
        path=path,
        headers=headers if headers else "",
        body=body if body else "",
        kwargs=kwargs if kwargs else "",
        nl0="\n" if kwargs else "",
        nl1="\n" if headers else "",
        nl2="\n" if body else "")


class MockUpResponse(CaosDBHTTPResponse):
    """A class for the responses of a request to the MockUpServerConnection."""

    def __init__(self, status, headers, body):
        self._status = status
        self.headers = headers
        self.response = StringIO(body)

    @property
    def status(self):
        """Return the status of the response."""
        return self._status

    def read(self, size=-1):
        """Return the body of the response."""
        return self.response.read(size).encode()

    def getheader(self, name, default=None):
        """Get the contents of the header `name`, or `default` if there is no
        matching header."""
        return self.headers[name] if name in self.headers else default

    def getheaders(self):
        return self.headers

    def close(self):
        pass


class MockUpServerConnection(CaosDBServerConnection):
    """The mock-up connection which does not actually connect to anything but
    just returns predefined responses which mimic the LinkAhead server."""

    def __init__(self):
        self.resources = [self._login]

    def _login(self, method, path, headers, body):
        if method == "POST" and path == "login":
            return MockUpResponse(200,
                                  headers={"AuthToken":
                                           "mockup-auth-token"},
                                  body="")

    def configure(self, **kwargs):
        """This configure method does nothing."""

    def request(self, method, path, headers=None, body=None, **kwargs):
        """Search a resource in the `resources` list which is answering to the
        request and return the response or raise a RuntimeError."""
        for resource in self.resources:
            response = resource(
                method=method, path=path, headers=headers, body=body, **kwargs)
            if response:
                return response
        raise RuntimeError(
            "No response for this request - " +
            _request_log_message(method, path, headers, body, **kwargs))




# THIS FILE IS GENERATED FROM linkahead SETUP.PY
#
short_version = '0.13.1'
version = '0.13.1'
full_version = '0.13.1'
git_revision = 'aa250ad55f9a5b417c6111c97bc1c32522035186'
release = True

if not release:
    version = full_version



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#

"""LinkAhead Python bindings.

Tries to read from the inifile specified in the environment variable `PYCAOSDBINI` or
alternatively in `~/.pylinkahead.ini` upon import.  After that, the ini file `pylinkahead.ini` in
the current working directory will be read additionally, if it exists.

"""

from os import environ, getcwd
# Import of the connection function (which is used to connect to the DB):
from os.path import expanduser, join
from warnings import warn

# Import of convenience methods:
from . import apiutils
from .common import administration
from .common.datatype import (BOOLEAN, DATETIME, DOUBLE, FILE, INTEGER, LIST,
                              REFERENCE, TEXT)
# Import of the basic  API classes:
from .common.models import (ACL, ALL, FIX, NONE, OBLIGATORY, RECOMMENDED,
                            SUGGESTED, Container, DropOffBox, Entity, File,
                            Info, Message, Permissions, Property, Query,
                            QueryTemplate, Record, RecordType, delete,
                            execute_query, get_global_acl,
                            get_known_permissions, raise_errors)
from .common.state import State, Transition
from .configuration import _read_config_files, configure, get_config
from .connection.connection import configure_connection, get_connection
from .exceptions import *
from .utils.get_entity import (get_entity_by_id, get_entity_by_name,
                               get_entity_by_path)

try:
    from .version import version as __version__
except ModuleNotFoundError:
    version = "uninstalled"
    __version__ = version

_read_config_files()



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""keyring.

A CredentialsProvider which uses the external 'keyring' library to
retrieve the password.
"""

import sys
import importlib
from getpass import getpass
from ...exceptions import ConfigurationError
from .external_credentials_provider import ExternalCredentialsProvider
from .interface import CredentialsAuthenticator


def get_authentication_provider():
    """get_authentication_provider.

    Return an authenticator which uses plain text username/password credentials.
    The difference to the `plain` module is that this implementation retrieves
    the password from the external gnome keyring.

    Returns
    -------
    CredentialsAuthenticator
        with a 'KeyringCaller' as back-end.
    """
    return CredentialsAuthenticator(KeyringCaller(callback=_call_keyring))


def _get_external_keyring():
    try:
        return importlib.import_module("keyring")
    except ImportError:
        raise RuntimeError(
            "The keyring password method requires installation of the"
            "keyring python package. On linux with python < 3.5, "
            "this requires the installation of dbus-python as a "
            "system package.")


def _call_keyring(**config):
    if "username" not in config:
        raise ConfigurationError("Your configuration did not provide a "
                                 "`username` which is needed by the "
                                 "`KeyringCaller` to retrieve the "
                                 "password in question.")
    url = config.get("url")
    username = config.get("username")
    app = "linkahead — {}".format(url)
    external_keyring = _get_external_keyring()
    password = external_keyring.get_password(app, username)
    if password is None:
        print("No password for user {} on {} found in keyring."
              .format(username, app))
        password = getpass("Enter password to save "
                           "in system keyring/wallet: ")
        external_keyring.set_password(app, username, password)
    return password


class KeyringCaller(ExternalCredentialsProvider):
    """KeyringCaller.

    A class for retrieving the password from the external 'gnome keyring' and
    storing the username/password credentials as plain text strings.

    Methods
    -------
    configure

    Attributes
    ----------
    password
    username
    """



# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2020 IndiScale GmbH <info@indiscale.com>
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header

import copy
from lxml import etree


def _translate_to_state_acis(acis):
    result = set()
    for aci in acis:
        aci = copy.copy(aci)
        if aci.role:
            aci.role = "?STATE?" + aci.role + "?"
        result.add(aci)
    return result


class Transition:
    """Transition

    Represents allowed transitions from one state to another.

    Properties
    ----------
    name : str
        The name of the transition
    description: str
        The description of the transition
    from_state : str
        A state name
    to_state : str
        A state name
    """

    def __init__(self, name, from_state, to_state, description=None):
        self._name = name
        self._from_state = from_state
        self._to_state = to_state
        self._description = description

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def from_state(self):
        return self._from_state

    @property
    def to_state(self):
        return self._to_state

    def __repr__(self):
        return f'Transition(name="{self.name}", from_state="{self.from_state}", to_state="{self.to_state}", description="{self.description}")'

    def __eq__(self, other):
        return (isinstance(other, Transition)
                and other.name == self.name
                and other.to_state == self.to_state
                and other.from_state == self.from_state)

    def __hash__(self):
        return 23472 + hash(self.name) + hash(self.from_state) + hash(self.to_state)

    @staticmethod
    def from_xml(xml):
        to_state = [to.get("name") for to in xml
                    if to.tag.lower() == "tostate"]
        from_state = [from_.get("name") for from_ in xml
                      if from_.tag.lower() == "fromstate"]
        result = Transition(name=xml.get("name"),
                            description=xml.get("description"),
                            from_state=from_state[0] if from_state else None,
                            to_state=to_state[0] if to_state else None)
        return result


class State:
    """State

    Represents the state of an entity and take care of the serialization and
    deserialization of xml for the entity state.

    An entity state is always a State of a StateModel.

    Properties
    ----------
    name : str
        Name of the State
    model : str
        Name of the StateModel
    description : str
        Description of the State (read-only)
    id : str
        Id of the undelying State record (read-only)
    transitions : set of Transition
        All transitions which are available from this state (read-only)
    """

    def __init__(self, model, name):
        self.name = name
        self.model = model
        self._id = None
        self._description = None
        self._transitions = None

    @property
    def id(self):
        return self._id

    @property
    def description(self):
        return self._description

    @property
    def transitions(self):
        return self._transitions

    def __eq__(self, other):
        return (isinstance(other, State)
                and self.name == other.name
                and self.model == other.model)

    def __hash__(self):
        return hash(self.name) + hash(self.model)

    def __repr__(self):
        return f"State('{self.model}', '{self.name}')"

    def to_xml(self):
        """Serialize this State to xml.

        Returns
        -------
        xml : etree.Element
        """
        xml = etree.Element("State")
        if self.name is not None:
            xml.set("name", self.name)
        if self.model is not None:
            xml.set("model", self.model)
        return xml

    @staticmethod
    def from_xml(xml):
        """Create a new State instance from an xml Element.

        Parameters
        ----------
        xml : etree.Element

        Returns
        -------
        state : State
        """
        name = xml.get("name")
        model = xml.get("model")
        result = State(name=name, model=model)
        result._id = xml.get("id")
        result._description = xml.get("description")
        transitions = [Transition.from_xml(t) for t in xml if t.tag.lower() ==
                       "transition"]
        if transitions:
            result._transitions = set(transitions)

        return result

    @staticmethod
    def create_state_acl(acl):
        from .models import ACL
        state_acl = ACL()
        state_acl._grants = _translate_to_state_acis(acl._grants)
        state_acl._denials = _translate_to_state_acis(acl._denials)
        state_acl._priority_grants = _translate_to_state_acis(acl._priority_grants)
        state_acl._priority_denials = _translate_to_state_acis(acl._priority_denials)
        return state_acl



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
# Copyright (C) 2020 IndiScale GmbH <info@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""server_side_scripting

Helper functions for calling server-side scripts.
"""
from urllib.parse import quote
from lxml import etree

from linkahead.connection.connection import get_connection
from linkahead.connection.utils import urlencode
from linkahead.connection.encode import (MultipartParam, multipart_encode,
                                         ReadableMultiparts)


def _make_params(pos_args, opts):
    """Create and return option string components.

The return value is a dict with be something like `-O<key>`:`<value>` from `opts` and
`-p{0,1,2,3,...}`:`<value>` from `pos_args`.

    """
    result = {}
    for key, val in opts.items():
        result["-O{key}".format(key=key)] = str(val)
    for i, val in enumerate(pos_args):
        result["-p{i}".format(i=i)] = str(val)
    return result


def _make_multipart_request(call, pos_args, opts, files):
    """Return body and header for an HTTP request.
    """
    parts = list()
    params = _make_params(pos_args, opts)

    parts.append(MultipartParam("call", call))
    for key, val in params.items():
        parts.append(MultipartParam(key, val))

    for paramname, filename in files.items():
        parts.append(MultipartParam.from_file(paramname=paramname,
                                              filename=filename))

    body, headers = multipart_encode(parts)
    body = ReadableMultiparts(body)
    return body, headers


def _make_form_request(call, pos_args, opts):
    """Return URL from call and argumewnts, and headers for urlencoding."""
    form = dict()
    form["call"] = call

    params = _make_params(pos_args, opts)
    for key, val in params.items():
        form[key] = val

    headers = {}
    headers["Content-Type"] = "application/x-www-form-urlencoded"
    return urlencode(form), headers


def _make_request(call, pos_args, opts, files=None):
    """
    Multipart if with files, otherwise url-encoded.

    Return
    ------
    path_segments, body, headers
    """

    if files is not None:
        return _make_multipart_request(call, pos_args, opts, files)

    return _make_form_request(call, pos_args, opts)


def run_server_side_script(call, *args, files=None, **kwargs):
    """

    Return
    ------
    response : ScriptingResponse
    """
    body, headers = _make_request(call=call, pos_args=args,
                                  opts=kwargs, files=files)
    response = get_connection()._http_request(method="POST",
                                              path=quote("scripting"),
                                              body=body,
                                              headers=headers)
    xml = etree.parse(response)
    code = int(xml.xpath("/Response/script/@code")[0])
    call = xml.xpath("/Response/script/call")[0].text
    stdout = xml.xpath("/Response/script/stdout")[0].text
    stderr = xml.xpath("/Response/script/stderr")[0].text

    return ScriptingResponse(call=call,
                             code=code,
                             stdout=stdout,
                             stderr=stderr)


class ScriptingResponse():
    """ScriptingResponse

    A data class for the response of server-side scripting calls.

    Properties
    ----------
    code : int
        The return code of the script process.
    call : str
        The complete call of the script minus the absolute path and the
        auth_token.
    stdout : str
        The STDOUT of the script process.
    stderr : str
        The STDERR of the script process.

    """

    def __init__(self, call, code, stdout, stderr):
        self.call = call
        self.code = code
        self.stdout = stdout
        self.stderr = stderr




from linkahead.connection.authentication.keyring import *
from warnings import warn

warn(("CaosDB was renamed to LinkAhead. Please import this library as `import linkahead.connection.authentication.keyring`. Using the"
      " old name, starting with caosdb, is deprecated."), DeprecationWarning)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
import warnings
import sys

from linkahead.utils.linkahead_admin import main

if __name__ == "__main__":
    warnings.warn("caosdb_admin.py is deprecated. Please use linkahead_admin.py",
                  DeprecationWarning)
    sys.exit(main())



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (c) 2019 Daniel Hornung
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""Connection to a LinkAhead server."""
from __future__ import absolute_import, print_function, unicode_literals

import logging
import ssl
import sys
import warnings
from builtins import str  # pylint: disable=redefined-builtin
from errno import EPIPE as BrokenPipe
from socket import error as SocketError
from urllib.parse import quote, urlparse
from warnings import warn

from requests import Session as HTTPSession
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError as HTTPConnectionError
from urllib3.poolmanager import PoolManager

from ..configuration import get_config
from ..exceptions import (ConfigurationError, HTTPClientError,
                          HTTPForbiddenError, HTTPResourceNotFoundError,
                          HTTPServerError, HTTPURITooLongError,
                          LinkAheadConnectionError, LinkAheadException,
                          LoginFailedError)

try:
    from ..version import version
except ModuleNotFoundError:
    version = "uninstalled"

from pkg_resources import resource_filename

from .encode import MultipartYielder, ReadableMultiparts
from .interface import CaosDBHTTPResponse, CaosDBServerConnection
from .utils import make_uri_path, parse_url, urlencode

_LOGGER = logging.getLogger(__name__)


class _WrappedHTTPResponse(CaosDBHTTPResponse):

    def __init__(self, response):
        self.response = response
        self._generator = None
        self._buffer = b''
        self._stream_consumed = False

    @property
    def reason(self):
        return self.response.reason

    @property
    def status(self):
        return self.response.status_code

    def read(self, size=None):
        if self._stream_consumed is True:
            raise RuntimeError("Stream is consumed")

        if self._buffer is None:
            # the buffer has been drained in the previous call.
            self._stream_consumed = True
            return b''

        if self._generator is None and (size is None or size == 0):
            # return full content at once
            self._stream_consumed = True
            return self.response.content

        if len(self._buffer) >= size:
            # still enough bytes in the buffer
            result = chunk[:size]
            self._buffer = chunk[size:]
            return result

        if self._generator is None:
            # first call to this method
            if size is None or size == 0:
                size = 512
            self._generator = self.response.iter_content(size)

        try:
            # read new data into the buffer
            chunk = self._buffer + next(self._generator)
            result = chunk[:size]
            if len(result) == 0:
                self._stream_consumed = True
            self._buffer = chunk[size:]
            return result
        except StopIteration:
            # drain buffer
            result = self._buffer
            self._buffer = None
            return result

    def getheader(self, name, default=None):
        return self.response.headers[name] if name in self.response.headers else default

    def getheaders(self):
        return self.response.headers.items()

    def close(self):
        self.response.close()


class _SSLAdapter(HTTPAdapter):
    """Transport adapter that allows us to use different SSL versions."""

    def __init__(self, ssl_version):
        self.ssl_version = ssl_version
        super().__init__()

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(
            num_pools=connections, maxsize=maxsize,
            block=block, ssl_version=self.ssl_version)


class _DefaultCaosDBServerConnection(CaosDBServerConnection):
    """_DefaultCaosDBServerConnection.

    Methods
    -------
    configure
    request
    """

    def __init__(self):
        self._useragent = ("linkahead-pylib/{version} - {implementation}".format(
            version=version, implementation=type(self).__name__))
        self._base_path = None
        self._session = None
        self._timeout = None

    def request(self, method, path, headers=None, body=None):
        """request.

        Send a HTTP request to the server.

        Parameters
        ----------
        method : str
            The HTTP request method.
        path : str
            An URI path segment (without the 'scheme://host:port/' parts),
            including query and frament segments.
        headers : dict of str -> str, optional
            HTTP request headers. (Defautl: None)
        body : str or bytes or readable, optional
            The body of the HTTP request. Bytes should be a utf-8 encoded
            string.

        Returns
        -------
        response : CaosDBHTTPResponse
        """

        if headers is None:
            headers = {}
        headers["User-Agent"] = self._useragent

        if path.endswith("/."):
            path = path[:-1] + "%2E"

        if isinstance(body, MultipartYielder):
            body = ReadableMultiparts(body)

        try:
            response = self._session.request(
                method=method,
                url=self._base_path + path,
                headers=headers,
                data=body,
                timeout=self._timeout,
                stream=True)

            return _WrappedHTTPResponse(response)
        except HTTPConnectionError as conn_err:
            raise LinkAheadConnectionError(
                "Connection failed. Network or server down? " + str(conn_err)
            )

    def configure(self, **config):
        """configure.

        Configure the http connection.

        Parameters
        ----------
        cacert : str
            Path to the CA certificate which will be used to identify the
            server.
        url : str
            The url of the LinkAhead Server, e.g.
            `https://example.com:443/rootpath`, including a possible root path.
        **config :
            Any further keyword arguments are being ignored.

        Raises
        ------
        LinkAheadConnectionError
            If no url has been specified, or if the CA certificate cannot be
            loaded.
        """

        if "url" not in config:
            raise LinkAheadConnectionError(
                "No connection url specified. Please "
                "do so via linkahead.configure_connection(...) or in a config "
                "file.")
        if (not config["url"].lower().startswith("https://") and not config["url"].lower().startswith("http://")):
            raise LinkAheadConnectionError("The connection url is expected "
                                           "to be a http or https url and "
                                           "must include the url scheme "
                                           "(i.e. start with https:// or "
                                           "http://).")

        url = urlparse(config["url"])
        path = url.path.strip("/")
        if len(path) > 0:
            path = path + "/"
        self._base_path = url.scheme + "://" + url.netloc + "/" + path

        self._session = HTTPSession()

        if url.scheme == "https":
            self._setup_ssl(config)

        # TODO(tf) remove in next release
        socket_proxy = config["socket_proxy"] if "socket_proxy" in config else None
        if socket_proxy is not None:
            self._session.proxies = {
                "https": "socks5://" + socket_proxy,
                "http": "socks5://" + socket_proxy,
            }

        if "https_proxy" in config:
            if self._session.proxies is None:
                self._session.proxies = {}
            self._session.proxies["https"] = config["https_proxy"]

        if "http_proxy" in config:
            if self._session.proxies is None:
                self._session.proxies = {}
            self._session.proxies["http"] = config["http_proxy"]

        if "timeout" in config:
            self._timeout = config["timeout"]

    def _setup_ssl(self, config):
        if "ssl_version" in config and config["cacert"] is not None:
            ssl_version = getattr(ssl, config["ssl_version"])
        else:
            ssl_version = ssl.PROTOCOL_TLS

        self._session.mount(self._base_path, _SSLAdapter(ssl_version))

        verify = True
        if "cacert" in config:
            verify = config["cacert"]
        if "ssl_insecure" in config and config["ssl_insecure"]:
            _LOGGER.warning("*** Warning! ***\n"
                            "Insecure SSL mode, certificate will not be checked! "
                            "Please consider removing the `ssl_insecure` configuration option.\n"
                            "****************")
            warnings.filterwarnings(action="ignore", module="urllib3",
                                    message="Unverified HTTPS request is being made")
            verify = False
        if verify is not None:
            self._session.verify = verify


def _make_conf(*conf):
    """_make_conf.

    Merge several config dicts into one. The precedence goes to latter dicts in
    the function call.

    Parameters
    ----------
    *conf : dict
        One ore more dicts with lower case option names (i.e. keys).

    Returns
    -------
    dict
        A merged config dict.
    """
    result = {}

    for conf_dict in conf:
        result.update(conf_dict)

    return result


_DEFAULT_CONF = {
    "password_method": "input",
    "implementation": _DefaultCaosDBServerConnection,
    "timeout": 210,
}


def _get_authenticator(**config):
    """_get_authenticator.

    Import and configure the password_method.

    Parameters
    ----------
    password_method : str
        The simple name of a submodule of linkahead.connection.authentication.
        Currently, there are four valid values for this parameter: 'plain',
        'pass', 'keyring' and 'auth_token'.
    **config :
        Any other keyword arguments are passed the configre method of the
        password_method.

    Returns
    -------
    AbstractAuthenticator
        An object which implements the password_method and which already
        configured.

    Raises
    ------
    ConfigurationError
        If the password_method string cannot be resolved to a LinkaheadAuthenticator
        class.
    """
    auth_module = ("linkahead.connection.authentication." +
                   config["password_method"])
    _LOGGER.debug("import auth_module %s", auth_module)
    try:
        __import__(auth_module)

        auth_provider = sys.modules[auth_module].get_authentication_provider()
        auth_provider.configure(**config)

        return auth_provider

    except ImportError:
        raise ConfigurationError("Password method \"{}\" not implemented. "
                                 "Try `plain`, `pass`, `keyring`, or "
                                 "`auth_token`."
                                 .format(config["password_method"]))


def configure_connection(**kwargs):
    """Configures the LinkAhead connection and returns the Connection object.

    The effective configuration is governed by the default values (see
    'Parameters'), the global configuration (see `linkahead.get_config()`) and the
    parameters which are passed to this function, with ascending priority.

    The parameters which are listed here, are possibly not sufficient for a
    working configuration of the connection. Check the `configure` method of
    the implementation class and the password_method for more details.

    Parameters
    ----------
    url : str
        The url of the LinkAhead Server. HTTP and HTTPS urls are allowed. However,
        it is **highly** recommend to avoid HTTP because passwords and
        authentication token are send over the network in plain text.

    username : str
        Username for login; e.g. 'admin'.

    password : str
        Password for login if 'plain' is used as password_method.

    password_method : str
        The name of a submodule of linkahead.connection.authentication which
        implements the AbstractAuthenticator interface. (Default: 'plain')
        Possible values are, for example:
        - "plain"    Need username and password arguments.
        - "input"    Asks for the password.
        - "pass"     Uses the `pass` password manager.
        - "keyring"  Uses the `keyring` library.
        - "auth_token" Uses only a given auth_token.

    timeout : int
        A connection timeout in seconds. (Default: 210)

    ssl_insecure : bool
        Whether SSL certificate warnings should be ignored. Only use this for
        development purposes! (Default: False)

    auth_token : str (optional)
        An authentication token which has been issued by the LinkAhead Server.
        Implies `password_method="auth_token"` if set.  An example token string would be `["O","OneTimeAuthenticationToken","anonymous",["administration"],[],1592995200000,604800000,"3ZZ4WKRB-5I7DG2Q6-ZZE6T64P-VQ","197d0d081615c52dc18fb323c300d7be077beaad4020773bb58920b55023fa6ee49355e35754a4277b9ac525c882bcd3a22e7227ba36dfcbbdbf8f15f19d1ee9",1,30000]`.

    https_proxy : str, optional
        Define a proxy for the https connections, e.g. `http://localhost:8888`,
        `socks5://localhost:8888`, or `socks4://localhost:8888`. These are
        either (non-TLS) HTTP proxies, SOCKS4 proxies, or SOCKS5 proxies. HTTPS
        proxies are not supported. However, the connection will be secured
        using TLS in the tunneled connection nonetheless. Only the connection
        to the proxy is insecure which is why it is not recommended to use HTTP
        proxies when authentication against the proxy is necessary. If
        unspecified, the https_proxy option of the pylinkahead.ini or the HTTPS_PROXY
        environment variable are being used. Use `None` to override these
        options with a no-proxy setting.

    http_proxy : str, optional
        Define a proxy for the http connections, e.g. `http://localhost:8888`.
        If unspecified, the http_proxy option of the pylinkahead.ini or the
        HTTP_PROXY environment variable are being used. Use `None` to override
        these options with a no-proxy setting.

    implementation : CaosDBServerConnection
        The class which implements the connection. (Default:
        _DefaultCaosDBServerConnection)

    Returns
    -------
    _Connection
        The singleton instance of the _Connection class.
    """
    global_conf = {}
    conf = get_config()
    # Convert config to dict, with preserving types
    int_opts = ["timeout"]
    bool_opts = ["ssl_insecure"]

    if conf.has_section("Connection"):
        global_conf = dict(conf.items("Connection"))
        # Integer options

        for opt in int_opts:
            if opt in global_conf:
                global_conf[opt] = conf.getint("Connection", opt)
        # Boolean options

        for opt in bool_opts:
            if opt in global_conf:
                global_conf[opt] = conf.getboolean("Connection", opt)
    local_conf = _make_conf(_DEFAULT_CONF, global_conf, kwargs)

    connection = _Connection.get_instance()

    if "socket_proxy" in local_conf:
        warnings.warn("Deprecated configuration option: socket_proxy. Use "
                      "the new https_proxy option instead",
                      DeprecationWarning, stacklevel=1)
    connection.configure(**local_conf)

    return connection


def get_connection():
    """Return the connection.

    If the connection was not configured yet `configure_connection` will
    be called inside this function without arguments.
    """
    connection = _Connection.get_instance()

    if connection.is_configured:
        return connection

    return configure_connection()


def _handle_response_status(http_response):

    status = http_response.status

    if status == 200:
        return

    # emtpy response buffer
    body = http_response.read()

    if status == 404:
        raise HTTPResourceNotFoundError("This resource has not been found.")
    elif status > 499:
        raise HTTPServerError(body=body)

    reason = http_response.reason
    standard_message = ("Request failed. The response returned with status "
                        "{} - {}.".format(status, reason))
    if status == 401:
        raise LoginFailedError(standard_message)
    elif status == 403:
        raise HTTPForbiddenError(standard_message)
    elif status in (413, 414):
        raise HTTPURITooLongError(standard_message)
    elif 399 < status < 500:
        raise HTTPClientError(msg=standard_message, status=status, body=body)
    else:
        raise LinkAheadException(standard_message)


class _Connection(object):  # pylint: disable=useless-object-inheritance
    """This connection class provides the interface to the database connection
    allowing for retrieval, insertion, update, etc. of entities, files, users,
    roles and much more.

    It wrapps an instance of CaosDBServerConnection which actually does the
    work (how, depends on the instance).

    It is a singleton and should not be instanciated or modified by any client.
    Use the methods `get_connection` and `configure_connection` for this
    purpose.
    """

    __instance = None

    def __init__(self):
        self._delegate_connection = None
        self._authenticator = None
        self.is_configured = False

    @classmethod
    def get_instance(cls):
        if cls.__instance is None:
            cls.__instance = _Connection()

        return cls.__instance

    def configure(self, **config):
        self.is_configured = True

        if "implementation" not in config:
            raise ConfigurationError(
                "Missing CaosDBServerConnection implementation. You did not "
                "specify an `implementation` for the connection.")
        try:
            self._delegate_connection = config["implementation"]()

            if not isinstance(self._delegate_connection,
                              CaosDBServerConnection):
                raise TypeError("The `implementation` callable did not return "
                                "an instance of CaosDBServerConnection.")
        except TypeError as type_err:
            raise ConfigurationError(
                "Bad CaosDBServerConnection implementation. The "
                "implementation must be a callable object which returns an "
                "instance of `CaosDBServerConnection` (e.g. a constructor "
                "or a factory).\n{}".format(type_err.args[0]))
        self._delegate_connection.configure(**config)

        if "auth_token" in config:
            # deprecated, needed for older scripts
            config["password_method"] = "auth_token"
        if "password_method" not in config:
            raise ConfigurationError("Missing password_method. You did "
                                     "not specify a `password_method` for"
                                     "the connection.")
        self._authenticator = _get_authenticator(
            connection=self._delegate_connection, **config)

        return self

    def retrieve(self, entity_uri_segments=None, query_dict=None, **kwargs):
        path = make_uri_path(entity_uri_segments, query_dict)

        http_response = self._http_request(method="GET", path=path, **kwargs)

        return http_response

    def delete(self, entity_uri_segments=None, query_dict=None, **kwargs):
        path = make_uri_path(entity_uri_segments, query_dict)

        http_response = self._http_request(
            method="DELETE", path=path, **kwargs)

        return http_response

    def update(self, entity_uri_segment, query_dict=None, **kwargs):
        path = make_uri_path(entity_uri_segment, query_dict)

        http_response = self._http_request(method="PUT", path=path, **kwargs)

        return http_response

    def activate_user(self, link):
        self._authenticator.logout()
        fullurl = urlparse(link)
        path = fullurl.path
        query = fullurl.query
        http_response = self._http_request(
            method="GET", path=path + "?" + query)

        return http_response

    def put_form_data(self, entity_uri_segment, params):
        return self._form_data_request(
            method="PUT", path=entity_uri_segment, params=params)

    def post_form_data(self, entity_uri_segment, params):
        return self._form_data_request(
            method="POST",
            path=entity_uri_segment,
            params=params)

    def _form_data_request(self, method, path, params):
        body = urlencode(params)
        headers = {}
        headers["Content-Type"] = "application/x-www-form-urlencoded"
        response = self._http_request(
            method=method,
            path=quote(path),
            body=body,
            headers=headers)

        return response

    def insert(self, entity_uri_segment, query_dict=None, body=None, **kwargs):
        path = make_uri_path(entity_uri_segment, query_dict)

        http_response = self._http_request(
            method="POST", path=path, body=body, **kwargs)

        return http_response

    def download_file(self, path):
        """This function downloads a file via HTTP from the LinkAhead file
        system."""
        try:
            uri_segments = ["FileSystem"]
            uri_segments.extend(path.split("/"))

            return self.retrieve(entity_uri_segments=uri_segments)
        except HTTPResourceNotFoundError:
            raise HTTPResourceNotFoundError("This file does not exist.")

    def _login(self):
        self._authenticator.login()

    def _logout(self):
        self._authenticator.logout()

    def _http_request(self, method, path, headers=None, body=None, **kwargs):
        try:
            return self._retry_http_request(method=method, path=path,
                                            headers=headers, body=body,
                                            **kwargs)
        except SocketError as e:
            if e.errno != BrokenPipe:
                raise

            return self._retry_http_request(method=method, path=path,
                                            headers=headers, body=body,
                                            reconnect=False,
                                            **kwargs)
        except LoginFailedError:
            if kwargs.get("reconnect", True) is True:
                self._login()

                return self._retry_http_request(method=method, path=path,
                                                headers=headers, body=body,
                                                reconnect=False,
                                                **kwargs)
            raise

    def _retry_http_request(self, method, path, headers, body, **kwargs):

        if hasattr(body, "encode"):
            # python3
            body = body.encode("utf-8")

        if headers is None:
            headers = {}
        self._authenticator.on_request(method=method, path=path,
                                       headers=headers)
        _LOGGER.debug("request: %s %s %s", method, path, str(headers))
        http_response = self._delegate_connection.request(
            method=method,
            path=path,
            headers=headers,
            body=body)
        _LOGGER.debug("response: %s %s", str(http_response.status),
                      str(http_response.getheaders()))
        self._authenticator.on_response(http_response)
        _handle_response_status(http_response)

        return http_response

    def get_username(self):
        """
        Return the username of the current connection.

        Shortcut for: get_connection()._authenticator._credentials_provider.username
        """
        return self._authenticator._credentials_provider.username



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
# Copyright (C) 2020 IndiScale GmbH <info@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
""" Versioning module for anything related to entity versions.

Currently this module defines nothing but a single class, `Version`.
"""

from __future__ import absolute_import
from .utils import xml2str
from lxml import etree


class Version():
    """The version of an entity.

    An entity version has a version id (string), a date (UTC timestamp), a
    list of predecessors and a list of successors.

    Parameters
    ----------
    id : str, optional
        See attribute `id`. Default: None
    date : str, optional
        See attribute `date`. Default: None
    username : str, optional
        See attribute `username`. Default: None
    realm : str, optional
        See attribute `realm`. Default: None
    predecessors : list of Version, optional
        See attribute `predecessors`. Default: empty list.
    successors : list of Version, optional
        See attribute `successors`. Default: empty list.
    is_head : bool
        See attribute `is_head`. Default: False
    is_complete_history : bool
        See attribute `is_complete_history`. Default: False

    Attributes
    ----------
    id : str
        Version ID (not the entity's id).
    date : str
        UTC Timestamp of the version, i.e. the date and time when the entity of
        this version has been inserted or modified.
    username : str
        The username of the user who inserted or updated this version.
    realm : str
        The realm of the user who inserted or updated this version.
    predecessors : list of Version
        Predecessors are the older entity versions which have been modified
        into this version. Usually, there is only one predecessor. However,
        this API allows for entities to be merged into one entity, which would
        result in more than one predecessor.
    successors : list of Version
        Successors are newer versions of this entity. If there are successors,
        this version is not the latest version of this entity. Usually, there
        is only one successor. However, this API allows that a single entity
        may co-exist in several versions (e.g. several proposals for the next
        entity status). That would result in more than one successor.
    is_head : bool or string
        If true, this indicates that this version is the HEAD if true.
        Otherwise it is not known whether this is the head or not.  Any string
        matching "true" (case-insensitively) is regarded as True.
        Nota bene: This property should typically be set if the server response
        indicated that this is the head version.
    is_complete_history : bool or string
        If true, this indicates that this version contains the full version
        history. That means, that the predecessors and successors have their
        respective predecessors and successors attached as well and the tree is
        completely available.  Any string matching "true" (case-insensitively)
        is regarded as True.
        Nota bene: This property should typically be set if the server response
        indicated that the full version history is included in its response.
    """

    # pylint: disable=redefined-builtin
    def __init__(self, id=None, date=None, username=None, realm=None,
                 predecessors=None, successors=None, is_head=False,
                 is_complete_history=False):
        """Typically the `predecessors` or `successors` should not "link back" to an existing Version
object."""
        self.id = id
        self.date = date
        self.username = username
        self.realm = realm
        self.predecessors = predecessors if predecessors is not None else []
        self.successors = successors if successors is not None else []
        self.is_head = str(is_head).lower() == "true"
        self.is_complete_history = str(is_complete_history).lower() == "true"

    def get_history(self):
        """ Returns a flat list of Version instances representing the history
        of the entity.

        The list items are ordered by the relation between the versions,
        starting with the oldest version.

        The items in the list have no predecessors or successors attached.

        Note: This method only returns reliable results if
        `self.is_complete_history is True` and it will not retrieve the full
        version history if it is not present.

        Returns
        -------
        list of Version
        """
        versions = []
        for p in self.predecessors:
            # assuming that predecessors don't have any successors
            versions = p.get_history()
        versions.append(Version(id=self.id, date=self.date,
                                username=self.username, realm=self.realm))
        for s in self.successors:
            # assuming that successors don't have any predecessors
            versions.extend(s.get_history())
        return versions

    def to_xml(self, tag="Version"):
        """Serialize this version to xml.

        The tag name is 'Version' per default. But since this method is called
        recursively for the predecessors and successors as well, the tag name
        can be configured.

        The resulting xml element contains attributes 'id' and 'date' and
        'Predecessor' and 'Successor' child elements.

        Parameters
        ----------
        tag : str, optional
            The name of the returned xml element. Defaults to 'Version'.

        Returns
        -------
        xml : etree.Element
        """
        xml = etree.Element(tag)
        if self.id is not None:
            xml.set("id", self.id)
        if self.date is not None:
            xml.set("date", self.date)
        if self.username is not None:
            xml.set("username", self.username)
        if self.realm is not None:
            xml.set("realm", self.realm)
        if self.predecessors is not None:
            for p in self.predecessors:
                xml.append(p.to_xml(tag="Predecessor"))
        if self.is_head is True:
            xml.set("head", "true")
        if self.successors is not None:
            for s in self.successors:
                xml.append(s.to_xml(tag="Successor"))
        return xml

    def __str__(self):
        """Return a stringified xml representation."""
        return self.__repr__()

    def __repr__(self):
        """Return a stringified xml representation."""
        return xml2str(self.to_xml())

    @staticmethod
    def from_xml(xml):
        """Parse a version object from a 'Version' xml element.

        Parameters
        ----------
        xml : etree.Element
            A 'Version' xml element, with 'id', possibly 'date', `username`,
            `realm`, and `head` attributes as well as 'Predecessor' and
            'Successor' child elements.

        Returns
        -------
        version : Version
            a new version instance
        """
        predecessors = [Version.from_xml(p) for p in xml if p.tag.lower() == "predecessor"]
        successors = [Version.from_xml(s) for s in xml if s.tag.lower() == "successor"]
        return Version(id=xml.get("id"), date=xml.get("date"),
                       is_head=xml.get("head"),
                       is_complete_history=xml.get("completeHistory"),
                       username=xml.get("username"), realm=xml.get("realm"),
                       predecessors=predecessors, successors=successors)

    def __hash__(self):
        """Hash of the version instance.

        Also hashes the predecessors and successors.
        """
        return (hash(self.id)
                + hash(self.date)
                + (Version._hash_list(self.predecessors)
                   if self.predecessors else 26335)
                + (Version._hash_list(self.successors)
                   if self.successors else -23432))

    @staticmethod
    def _hash_list(_list):
        """Hash a list by hashing each element and its index."""
        result = 12352
        for idx, val in enumerate(_list):
            result += hash(val) + idx
        return result

    @staticmethod
    def _eq_list(this, that):
        """List equality.

        List equality is defined as equality of each element, the order
        and length.
        """
        if len(this) != len(that):
            return False
        for v1, v2 in zip(this, that):
            if v1 != v2:
                return False
        return True

    def __eq__(self, other):
        """Equality of versions is defined by equality of id, date, and list
        equality of the predecessors and successors."""
        return (self.id == other.id
                and self.date == other.date
                and Version._eq_list(self.predecessors, other.predecessors)
                and Version._eq_list(self.successors, other.successors))



# -*- coding: utf-8 -*-
#
# This file is a part of the CaosDB Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2020 Timm Fitschen <t.fitschen@indiscale.com>
# Copyright (C) 2020-2022 IndiScale GmbH <info@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""git-utils: Some functions for retrieving information about git repositories.

"""

import logging
import tempfile

from subprocess import call

logger = logging.getLogger(__name__)


def get_origin_url_in(folder: str):
    """return the Fetch URL of the git repository in the given folder."""
    with tempfile.NamedTemporaryFile(delete=False, mode="w") as t:
        call(["git", "remote", "show", "origin"], stdout=t, cwd=folder)
    with open(t.name, "r") as t:
        urlString = "Fetch URL:"

        for line in t.readlines():
            if urlString in line:
                return line[line.find(urlString) + len(urlString):].strip()

    return None


def get_diff_in(folder: str, save_dir=None):
    """returns the name of a file where the out put of "git diff" in the given
    folder is stored."""
    with tempfile.NamedTemporaryFile(delete=False, mode="w", dir=save_dir) as t:
        call(["git", "diff"], stdout=t, cwd=folder)

    return t.name


def get_branch_in(folder: str):
    """returns the current branch of the git repository in the given folder.

    The command "git branch" is called in the given folder and the
    output is returned
    """
    with tempfile.NamedTemporaryFile(delete=False, mode="w") as t:
        call(["git", "rev-parse", "--abbrev-ref", "HEAD"], stdout=t, cwd=folder)
    with open(t.name, "r") as t:
        return t.readline().strip()


def get_commit_in(folder: str):
    """returns the commit hash in of the git repository in the given folder.

    The command "git log -1 --format=%h" is called in the given folder
    and the output is returned
    """

    with tempfile.NamedTemporaryFile(delete=False, mode="w") as t:
        call(["git", "log", "-1", "--format=%h"], stdout=t, cwd=folder)
    with open(t.name, "r") as t:
        return t.readline().strip()



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#

"""provides utilities for creating revisions of database entries."""
import linkahead as db
from linkahead import INTEGER, LIST


def bend_references(from_id, to_id, except_for=None):
    """redirects all references to a new target.

    All entities having a reference pointing to from_id are found
    and those references are changed to point to to_id.
    entities having an id listed in except_for are excluded.

Parameters
----------

from_id : int
  the old object to which references where pointing
to_id : int
  the new object to which references will be pointing
except_for : list of int
  entities with id of this list will not be changed
    """
    if except_for is None:
        except_for = [to_id]
    entities = db.execute_query(
        "FIND ENTITY WHICH references {}".format(from_id))
    for ent in entities:
        if ent.id in except_for:
            continue
        for prop in ent.properties:
            if isinstance(
                    prop.value, int) and (
                        prop.datatype != INTEGER) and (
                            prop.value == from_id):
                prop.value = to_id
            if (isinstance(prop.value, list)
                    and len(prop.value) > 0
                    and isinstance(prop.value[0], int)
                    and (prop.datatype != LIST(INTEGER))
                    and from_id in prop.value):
                index = prop.value.index(from_id)
                prop.value[index] = to_id
        ent.update()


def create_revision(old_id, prop, value):
    """creates a revision of an existing record.

    This function changes the record with id old_id. The value of the
    propertye prop is changed to value.

Parameters
----------

old_id : int
    id of the record to be changed
prop : string
    name of the property to be changed
value : type of corresponding property
    the new value of the corresponding property
"""
    record = db.execute_query("FIND {}".format(old_id))[0]
    new_rec = record.copy()
    new_rec.get_property(prop).value = value
    try:
        new_rec.remove_property("revisionOf")
    except BaseException:
        pass
    new_rec.add_property(name="revisionOf", value=record.id)
    new_rec.insert()
    bend_references(record.id, new_rec.id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#

"""A small LinkAhead client with a focus on administration of the server."""

from __future__ import print_function, unicode_literals

import getpass
import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter

import linkahead as db
from linkahead import administration as admin
from linkahead.exceptions import HTTPClientError

__all__ = []
__version__ = 0.3
__date__ = '2016-09-19'
__updated__ = '2018-12-11'


def do_update_role(args):
    admin._update_role(name=args.role_name, description=args.role_description)


def do_create_role(args):
    admin._insert_role(name=args.role_name, description=args.role_description)


def do_retrieve_role(args):
    print(admin._retrieve_role(name=args.role_name))


def do_delete_role(args):
    admin._delete_role(name=args.role_name)


def do_retrieve(args):
    c = None

    if args.query:
        if len(args.entities) > 1:
            raise Exception("Only one query at a time can be retrieved.")
        c = db.execute_query(args.entities[0], flags=eval(args.flags))
    else:
        c = db.Container()

        for i in args.entities:
            try:
                eid = int(i)
                c.append(db.Entity(id=eid))
            except ValueError:
                c.append(db.Entity(name=i))
        c.retrieve(flags=eval(args.flags))
    print(c)


def do_update(args):
    fdict = eval(args.flags)
    xml = open(args.xml_path, "r")
    ret = db.get_connection().update(
        entity_uri_segment=["Entity"], reconnect=True, body=xml)
    db.Container._response_to_entities(ret)


def do_delete(args):
    c = db.Container()

    for i in args.entities:
        c.append(db.Entity(id=i))

    c.delete()


def do_insert(args):
    fdict = eval(args.flags)
    xml = open(args.xml_path, "r")
    ret = db.get_connection().insert(
        entity_uri_segment=["Entity"],
        reconnect=True,
        query_dict=fdict,
        body=xml)
    print(db.Container._response_to_entities(ret))


def _promt_for_pw():
    password = getpass.getpass(prompt="Please type password: ")
    password2 = getpass.getpass(prompt="Please type password again: ")

    if password != password2:
        raise Exception("Password strings didn't match")

    return password


def do_create_user(args):
    password = args.user_password

    if args.ask_password is True:
        password = _promt_for_pw()
    try:
        admin._insert_user(name=args.user_name,
                           email=args.user_email, password=password)

        if args.activate_user:
            do_activate_user(args)
    except HTTPClientError as e:
        print(e.msg)


def do_activate_user(args):
    try:
        admin._update_user(name=args.user_name, status="ACTIVE")
    except HTTPClientError as e:
        print(e.msg)


def do_deactivate_user(args):
    try:
        admin._update_user(name=args.user_name, status="INACTIVE")
    except HTTPClientError as e:
        print(e.msg)


def do_set_user_password(args):
    if args.user_password is None:
        password = _promt_for_pw()
    else:
        password = args.user_password
    try:
        admin._update_user(name=args.user_name, password=password)
    except HTTPClientError as e:
        print(e.msg)


def do_add_user_roles(args):
    roles = admin._get_roles(username=args.user_name, realm=None)

    for r in args.user_roles:
        roles.add(r)
    admin._set_roles(username=args.user_name, roles=roles)


def do_remove_user_roles(args):
    roles = admin._get_roles(username=args.user_name, realm=None)

    for r in args.user_roles:
        if r in roles:
            roles.remove(r)
    admin._set_roles(username=args.user_name, roles=roles)


def do_set_user_entity(args):
    admin._update_user(name=args.user_name, entity=args.user_entity)


def do_reset_user_entity(args):
    admin._update_user(name=args.user_name, entity="")


def do_set_user_email(args):
    admin._update_user(name=args.user_name, email=args.user_email)


def do_retrieve_user(args):
    print(admin._retrieve_user(name=args.user_name))


def do_delete_user(args):
    admin._delete_user(name=args.user_name)


def do_retrieve_user_roles(args):
    print(admin._get_roles(username=args.user_name))


def do_retrieve_role_permissions(args):
    print(admin._get_permissions(role=args.role_name))


def do_grant_role_permissions(args):
    perms = admin._get_permissions(args.role_name)

    for p in args.role_permissions:
        g = admin.PermissionRule(
            action="Grant", permission=p, priority=args.permissions_priority)
        d = admin.PermissionRule(
            action="Deny", permission=p, priority=args.permissions_priority)

        if g in perms:
            perms.remove(g)

        if d in perms:
            perms.remove(d)
        perms.add(g)
    admin._set_permissions(role=args.role_name, permission_rules=perms)


def do_revoke_role_permissions(args):
    perms = admin._get_permissions(args.role_name)

    for p in args.role_permissions:
        g = admin.PermissionRule(
            action="Grant", permission=p, priority=args.permissions_priority)
        d = admin.PermissionRule(
            action="Deny", permission=p, priority=args.permissions_priority)

        if g in perms:
            perms.remove(g)

        if d in perms:
            perms.remove(d)
    admin._set_permissions(role=args.role_name, permission_rules=perms)


def do_deny_role_permissions(args):
    perms = admin._get_permissions(args.role_name)

    for p in args.role_permissions:
        g = admin.PermissionRule(
            action="Grant", permission=p, priority=args.permissions_priority)
        d = admin.PermissionRule(
            action="Deny", permission=p, priority=args.permissions_priority)

        if g in perms:
            perms.remove(g)

        if d in perms:
            perms.remove(d)
        perms.add(d)
    admin._set_permissions(role=args.role_name, permission_rules=perms)


def do_retrieve_entity_acl(args):
    entities = db.execute_query(q=args.query, flags={"ACL": None})

    for entity in entities:
        print(entity.id)
        print(entity.acl)


def do_action_entity_permissions(args):
    entities = db.execute_query(q=args.query, flags={"ACL": None})

    for entity in entities:
        for p in args.permissions:
            getattr(entity, args.action)(role=args.role, priority=args.priority,
                                         permission=p)
    entities.update(flags={"ACL": None})

    for entity in entities:
        print(entity.id)
        print(entity.acl)


def main(argv=None):
    """Command line options."""

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    # program_name = os.path.basename(sys.argv[0])
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = '%%(prog)s %s (%s)' % (
        program_version, program_build_date)
    program_shortdesc = __import__('__main__').__doc__
    program_license = '''%s

USAGE
''' % (program_shortdesc)

    # Setup argument parser
    parser = ArgumentParser(description=program_license,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('-V', '--version', action='version',
                        version=program_version_message)
    parser.add_argument("--auth-token", metavar="AUTH_TOKEN",
                        dest="auth_token",
                        help=("A LinkAhead authentication token (default: None). "
                              "If the authentication token is passed, the "
                              "`password_method` of the connection is set to "
                              "`auth_token` and the respective configuration "
                              "from the pylinkahead.ini is effectively being "
                              "overridden.\nTODO: Also allow passing the token "
                              "via environmenty variables."))
    subparsers = parser.add_subparsers(
        title="commands",
        metavar="COMMAND",
        description="You can invoke the following commands. Print the detailed help for each command with #> linkahead_admin COMMAND -h")

    # users (CRUD)
    subparser = subparsers.add_parser(
        "create_user",
        help="Create a new user in LinkAhead's internal user database. You need "
        " to activate the user before use.")
    subparser.set_defaults(call=do_create_user)
    mg = subparser.add_mutually_exclusive_group()
    mg.add_argument("-a", "--ask-password",
                    help="Prompt for a password.", action="store_true")
    mg.add_argument(
        "--password",
        dest="user_password",
        default=None,
        help="Alternative way to provide the new user's password. Please "
        "consider to use the more secure, interactive way (-a option).")
    subparser.add_argument("-c", "--activate-user",
                           help="Activate the user after creation.",
                           action="store_true")
    subparser.add_argument(
        metavar='USERNAME',
        dest="user_name",
        help="A user name which is unique in the internal user database.")
    subparser.add_argument(
        metavar="EMAIL",
        nargs='?',
        dest="user_email",
        help="The email address of the new user.")

    subparser = subparsers.add_parser(
        "activate_user", help="(Re-)activate an inactive (but existing) user.")
    subparser.set_defaults(call=do_activate_user)
    subparser.add_argument(metavar='USERNAME', dest="user_name",
                           help="The name of the user who is to be activated.")

    subparser = subparsers.add_parser(
        "deactivate_user", help="Deactivate an active user.")
    subparser.set_defaults(call=do_deactivate_user)
    subparser.add_argument(
        metavar='USERNAME',
        dest="user_name",
        help="The name of the user who is to be deactivated.")

    subparser = subparsers.add_parser(
        "set_user_password",
        help="Set a new password for a user. "
        "By default, you will be prompted for the password.")
    subparser.set_defaults(call=do_set_user_password)
    subparser.add_argument(
        metavar='USERNAME',
        dest="user_name",
        help="The name of the user who's password is to be set.")
    subparser.add_argument(
        metavar='PASSWORD',
        nargs="?",
        dest="user_password",
        default=None,
        help="Alternative way to provide the user's new password. "
        "The more secure (and default way) is to provide it interactively.")

    subparser = subparsers.add_parser(
        "set_user_entity",
        help="Associate a user with an existing entity (which should represent a person, a program, an organization or something similar).")
    subparser.set_defaults(call=do_set_user_entity)
    subparser.add_argument(
        metavar='USERNAME',
        dest="user_name",
        help="The name of the user who's associated entity you want to set.")
    subparser.add_argument(metavar='ENTITY', dest="user_entity",
                           help="An ID of an existing entity.")

    subparser = subparsers.add_parser(
        "reset_user_entity",
        help="Terminate the association of a user with an entity.")
    subparser.set_defaults(call=do_reset_user_entity)
    subparser.add_argument(
        metavar='USERNAME',
        dest="user_name",
        help="The name of the user who's associated entity you want to reset.")

    subparser = subparsers.add_parser(
        "set_user_email", help="Set a new email for a user.")
    subparser.set_defaults(call=do_set_user_email)
    subparser.add_argument(
        metavar='USERNAME',
        dest="user_name",
        help="The name of the user who's email is to be set.")
    subparser.add_argument(
        metavar='EMAIL',
        dest="user_email",
        help="The name of the user who's email is to be set.")

    subparser = subparsers.add_parser(
        "retrieve_user", help="Retrieve a user (email, entity)")
    subparser.set_defaults(call=do_retrieve_user)
    subparser.add_argument(
        metavar='USERNAME', dest="user_name", help="The name of the user.")

    subparser = subparsers.add_parser(
        "delete_user",
        help="Delete a user from linkahead's internal user database.")
    subparser.set_defaults(call=do_delete_user)
    subparser.add_argument(metavar='USERNAME', dest="user_name",
                           help="The name of the user who is to be deleted.")

    # user roles
    subparser = subparsers.add_parser(
        "add_user_roles", help="Extend the roles of a user.")
    subparser.set_defaults(call=do_add_user_roles)
    subparser.add_argument(
        metavar='USERNAME',
        dest="user_name",
        help="The name of the user who's roles are to be extended.")
    subparser.add_argument(
        metavar='ROLES',
        dest="user_roles",
        nargs='+',
        help="A space separated list of (existing) roles.")

    subparser = subparsers.add_parser(
        "remove_user_roles", help="Remove some of the roles of a user.")
    subparser.set_defaults(call=do_remove_user_roles)
    subparser.add_argument(
        metavar='USERNAME',
        dest="user_name",
        help="The name of the user from whom you want to take some roles away.")
    subparser.add_argument(
        metavar='ROLES',
        dest="user_roles",
        nargs='+',
        help="A space separated list of (existing) roles.")

    subparser = subparsers.add_parser(
        "retrieve_user_roles", help="Retrieve a user's roles.")
    subparser.set_defaults(call=do_retrieve_user_roles)
    subparser.add_argument(
        metavar='USERNAME', dest="user_name", help="The name of the user.")

    # role permissions
    subparser = subparsers.add_parser(
        "retrieve_role_permissions",
        help="Retrieve the set of permission rules of a role.")
    subparser.set_defaults(call=do_retrieve_role_permissions)
    subparser.add_argument(
        metavar='ROLE',
        dest="role_name",
        help="The name of the role which permissions are to be retrieved.")

    subparser = subparsers.add_parser(
        "grant_role_permissions", help="Grant permissions to a role.")
    subparser.set_defaults(call=do_grant_role_permissions)
    subparser.add_argument(
        '--priority',
        dest="permissions_priority",
        action="store_true",
        default=False,
        help="This flag enables priority permission rules.")
    subparser.add_argument(
        metavar='ROLE',
        dest="role_name",
        help="The name of the role to which the permissions are to be granted.")
    subparser.add_argument(
        metavar='PERMISSIONS',
        dest="role_permissions",
        nargs='+',
        help="A space separated list of permissions.")

    subparser = subparsers.add_parser(
        "revoke_role_permissions",
        help="Remove previously granted or denied permissions from a role.")
    subparser.set_defaults(call=do_revoke_role_permissions)
    subparser.add_argument(
        '--priority',
        dest="permissions_priority",
        action="store_true",
        default=False,
        help="This flag is needed to revoke priority permissions.")
    subparser.add_argument(
        metavar='ROLE',
        dest="role_name",
        help="The name of the role from which you want to revoke permissions.")
    subparser.add_argument(
        metavar='PERMISSIONS',
        dest="role_permissions",
        nargs='+',
        help="A space separated list of permissions.")

    subparser = subparsers.add_parser(
        "deny_role_permissions", help="Deny a role permissions.")
    subparser.set_defaults(call=do_deny_role_permissions)
    subparser.add_argument(
        '--priority',
        dest="permissions_priority",
        action="store_true",
        default=False,
        help="This flag enables priority permission rules.")
    subparser.add_argument(
        metavar='ROLE',
        dest="role_name",
        help="The name of the role which you want to deny permissions.")
    subparser.add_argument(
        metavar='PERMISSIONS',
        dest="role_permissions",
        nargs='+',
        help="A space separated list of permissions.")

    # entities (CRUD)
    subparser = subparsers.add_parser("insert", help="Insert entities.")
    subparser.set_defaults(call=do_insert)
    subparser.add_argument(
        '-f',
        '--flags',
        dest="flags",
        help="A python dictionary (dict) with flag keys and their values.",
        metavar="FLAGS",
        default="{}")
    subparser.add_argument(metavar='PATH', dest="xml_path",
                           help="Path to an xml file.")

    subparser = subparsers.add_parser("retrieve", help="Retrieve entities.")
    subparser.set_defaults(call=do_retrieve)
    subparser.add_argument(
        '-f',
        '--flags',
        dest="flags",
        help="A python dictionary (dict) with flag keys and their values.",
        metavar="FLAGS",
        default="{}")
    subparser.add_argument('-q', '--query', dest='query', action="store_true",
                           help="If the ENTITIES argument is a query.")
    subparser.add_argument(metavar='ENTITIES', dest="entities", nargs='+',
                           help="A space separated list of ids or names of"
                           "entities or ai single query.")

    subparser = subparsers.add_parser("update", help="Update entities.")
    subparser.set_defaults(call=do_update)
    subparser.add_argument(
        '-f',
        '--flags',
        dest="flags",
        help="A python dictionary (dict) with flag keys and their values.",
        metavar="FLAGS",
        default="{}")
    subparser.add_argument(metavar='PATH', dest="xml_path",
                           help="Path to an xml file.")

    subparser = subparsers.add_parser("delete", help="Delete entities.")
    subparser.set_defaults(call=do_delete)
    subparser.add_argument(
        '-f',
        '--flags',
        dest="flags",
        help="A python dictionary (dict) with flag keys and their values.",
        metavar="FLAGS",
        default="{}")
    subparser.add_argument(
        metavar='ENTITIES',
        dest="entities",
        nargs='+',
        help="A space separated list of ids or names of entities.")

    # roles (CRUD)
    create_role_parser = subparsers.add_parser(
        "create_role", help="Create a new role.")
    create_role_parser.set_defaults(call=do_create_role)
    create_role_parser.add_argument(
        dest="role_name", metavar="ROLENAME", help="The name of the new role.")
    create_role_parser.add_argument(
        dest="role_description",
        metavar="DESCRIPTION",
        help="A description of the role's purpose, it's intended use case, characteristics of the users who have this role, etc.")

    retrieve_role_parser = subparsers.add_parser(
        "retrieve_role", help="Retrieve the description of an existing role.")
    retrieve_role_parser.set_defaults(call=do_retrieve_role)
    retrieve_role_parser.add_argument(
        dest="role_name",
        metavar="ROLENAME",
        help="The name of the existing role.")

    update_role_parser = subparsers.add_parser(
        "update_role", help="Change the description of an existing role.")
    update_role_parser.set_defaults(call=do_update_role)
    update_role_parser.add_argument(
        dest="role_name",
        metavar="ROLENAME",
        help="The name of the existing role.")
    update_role_parser.add_argument(
        dest="role_description",
        metavar="DESCRIPTION",
        help="A new description of the role's purpose, it's intended use case, characteristics of the users who have this role, etc.")

    delete_role_parser = subparsers.add_parser(
        "delete_role", help="Delete a role.")
    delete_role_parser.set_defaults(call=do_delete_role)
    delete_role_parser.add_argument(
        dest="role_name",
        metavar="ROLENAME",
        help="The name of the existing role.")

    # entity acl
    retrieve_entity_acl_parser = subparsers.add_parser(
        "retrieve_entity_acl", help="Retrieve an entity ACL.")
    retrieve_entity_acl_parser.set_defaults(call=do_retrieve_entity_acl)
    retrieve_entity_acl_parser.add_argument(dest="query", metavar="QUERY",
                                            help="A FIND query.")

    for action in ["grant", "deny", "revoke_denial", "revoke_grant"]:
        action_entity_permissions_parser = subparsers.add_parser(
            f"{action}_entity_permissions",
            help=f"{action} entity permissions to one or more Entities.")
        action_entity_permissions_parser.set_defaults(
            call=do_action_entity_permissions, action=action)
        action_entity_permissions_parser.add_argument(dest="query", metavar="QUERY",
                                                      help="A FIND query.")
        action_entity_permissions_parser.add_argument(dest="role", metavar="ROLE",
                                                      help="The name of an exising role.")
        action_entity_permissions_parser.add_argument(
            dest="permissions",
            metavar="PERMISSION",
            help="A list of permissions",
            nargs='+')
        action_entity_permissions_parser.add_argument(
            '--priority',
            dest="priority",
            action="store_true",
            default=False,
            help="This flag enables priority permission rules.")

    # Process arguments
    args = parser.parse_args()
    auth_token = args.auth_token
    if auth_token is not None:
        db.configure_connection(password_method="auth_token",
                                auth_token=auth_token)
    else:
        db.configure_connection()

    return args.call(args)


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2020 IndiScale GmbH
# Copyright (C) 2020 Henrik tom Wörden, IndiScale GmbH
# Copyright (C) 2020 Daniel Hornung (d.hornung@indiscale.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#

import re

from ..exceptions import EmptyUniqueQueryError, QueryNotUniqueError

DOUBLE = "DOUBLE"
REFERENCE = "REFERENCE"
TEXT = "TEXT"
DATETIME = "DATETIME"
INTEGER = "INTEGER"
FILE = "FILE"
BOOLEAN = "BOOLEAN"


def LIST(datatype):
    if hasattr(datatype, "name"):
        datatype = datatype.name

    return "LIST<" + str(datatype) + ">"


def get_list_datatype(datatype):
    """ returns the datatype of the elements in the list """
    if not isinstance(datatype, str):
        return None
    match = re.match("LIST(<|&lt;)(?P<datatype>.*)(>|&gt;)", datatype)

    if match is not None:
        return match.group("datatype")
    else:
        return None


def is_list_datatype(datatype):
    """ returns whether the datatype is a list """

    return get_list_datatype(datatype) is not None


def is_reference(datatype):
    """Returns whether the value is a reference

    FILE and REFERENCE properties are examples, but also datatypes that are
    RecordTypes.

    Parameters
    ----------
    datatype : str
               The datatype to check.

    Returns
    -------
    bool
        True if the datatype is a not base datatype or a list of a base datatype.
        Otherwise False is returned.
    """

    if datatype is None:
        raise ValueError("Cannot decide whether datatype is reference if None"
                         " is supplied")

    if datatype in [DOUBLE, BOOLEAN, INTEGER, TEXT, DATETIME]:
        return False
    elif is_list_datatype(datatype):
        return is_reference(get_list_datatype(datatype))
    else:
        return True


def get_referenced_recordtype(datatype):
    """Return the record type of the referenced datatype.

    Raises
    ------
    ValueError
              In cases where datatype is not a reference, the list does not have
              a referenced record type or the datatype is a FILE.

    Parameters
    ----------
    datatype : str
               The datatype to check.

    Returns
    -------
    str
       String containing the name of the referenced datatype.
    """

    if not is_reference(datatype):
        raise ValueError("datatype must be a reference")

    if is_list_datatype(datatype):
        datatype = get_list_datatype(datatype)
        if datatype is None:
            raise ValueError("list does not have a list datatype")

    if datatype == FILE:
        raise ValueError(
            "FILE references are not considered references with a record type")

    return datatype


def get_id_of_datatype(datatype):
    """ returns the id of a Record Type

    This is not trivial, as queries may also return children. A check comparing
    names is necessary.

    Parameters
    ----------
    datatype : string
        A datatype, e.g. DOUBLE, or LIST<Person>

    Returns
    -------
    The id of the RecordType with the same name as the datatype.

    Raises
    ------
    QueryNotUniqueError
        If there are more than one entities with the same name as the datatype.
    EmptyUniqueQueryError
        If there is no entity with the name of the datatype.
    """

    from .models import execute_query
    if is_list_datatype(datatype):
        datatype = get_list_datatype(datatype)
    q = "FIND RECORDTYPE {}".format(datatype)

    # we cannot use unique=True here, because there might be subtypes
    res = execute_query(q)
    res = [el for el in res if el.name.lower() == datatype.lower()]

    if len(res) > 1:
        raise QueryNotUniqueError(
            "Name {} did not lead to unique result; Missing "
            "implementation".format(datatype))
    elif len(res) != 1:
        raise EmptyUniqueQueryError(
            "No RecordType named {}".format(datatype))

    return res[0].id



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#

"""!!! Deprecated !!! YAML interface for the database (LinkAhead)"""

import yaml
from lxml import etree
from lxml.etree import Element
import re
import linkahead
import linkahead.common.utils as utils
from linkahead.connection.connection import get_connection
import warnings


def append_sublist(v, newel, def_entity_type):
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use linkahead-advanced-user-tools/models/data_model.py for a
                  similar functionality.""", DeprecationWarning)
    if v is None:
        return
    for i in v:
        if isinstance(i, dict):
            entity_type = def_entity_type
            if "entity" in i:
                entity_type = i["entity"]
                del i["entity"]
            newel.append(kv_to_xml(entity_type, i))


def kv_to_xml(k, v):
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use linkahead-advanced-user-tools/models/data_model.py for a
                  similar functionality.""", DeprecationWarning)
    newel = Element(k)
    # code.interact(local=locals())
    if isinstance(v, list):  # Top level loop
        append_sublist(v, newel, "Property")
    elif isinstance(v, dict):
        if "value" in v:
            newel.text = v["value"]
            del v["value"]
        subdict = {"properties": "Property", "parents": "Parent"}
        for jk, jv in subdict.items():
            if jk in v:
                append_sublist(v[jk], newel, jv)
                del v[jk]
        for k2, v2 in v.items():
            newel.set(k2, str(v2))
    return newel


def dict_to_xml(d):
    """
    d: The dictionary (possibly loaded from yaml)
       to convert to linkahead-xml.
    """
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use linkahead-advanced-user-tools/models/data_model.py for a
                  similar functionality.""", DeprecationWarning)
    return kv_to_xml("Entities", d)


def yaml_to_xml(yamlstr):
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use linkahead-advanced-user-tools/models/data_model.py for a
                  similar functionality.""", DeprecationWarning)
    """Load a yaml document from yamlstr and converts it to XML.

    Parameters
    ----------
    yamlstr : str
        The string to load the yaml document from.

    """
    return dict_to_xml(yaml.load(yamlstr, Loader=yaml.SafeLoader))


def process(text):
    """Do some replacements on the original file to obtain valid yaml."""
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use linkahead-advanced-user-tools/models/data_model.py for a
                  similar functionality.""", DeprecationWarning)
    processed = re.sub(
        "^(\\s*)-\\s*\\{?(.*)\\}?\\s*$",
        "\\1- {\\2}",
        text,
        flags=re.MULTILINE)
    processed = re.sub("^(\\s*)\\+\\s*(.*)\\s*$", "\\1- \\2",
                       processed, flags=re.MULTILINE)
    print(processed)
    return processed


def yaml_file_to_xml(yamlfilename):
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use linkahead-advanced-user-tools/models/data_model.py for a
                  similar functionality.""", DeprecationWarning)
    with open(yamlfilename, "r") as f:
        return yaml_to_xml(process(f.read()))


def insert_yaml_file(yamlfilename, simulate=False):
    """Inserts the contents of 'yamlfilename' into the database.

    Set 'simulate' to True if you don't actually want to insert the xml,
    but only receive what would be sent.
    """
    warnings.warn("""
                  This function is deprecated and will be removed with the next release.
                  Please use linkahead-advanced-user-tools/models/data_model.py for a
                  similar functionality.""", DeprecationWarning)
    con = get_connection()
    prs = etree.XMLParser(remove_blank_text=True)
    sent_xml = etree.tostring(
        etree.fromstring(
            etree.tostring(
                yaml_file_to_xml(yamlfilename)),
            prs),
        pretty_print=True)
    if simulate:
        return "", sent_xml.decode("utf-8")
    response = con.insert(entity_uri_segment="Entity/",
                          body=sent_xml)
    resp_text = response.readall()
    resp_elem = etree.fromstring(resp_text, prs)
    for i in resp_elem.iter("Error"):
        print("ERROR: " + i.get("description"))
        child = i.getparent()
        while child is not None:
            childname = ""
            childid = ""
            # print(etree.tostring(child))
            if child.get("name") is not None:
                childname = child.get("name")
            if child.get("id") is not None:
                childid = child.get("id")
            print("  in " + child.tag + " " + childname + " " + childid)
            child = child.getparent()
    return etree.tostring(resp_elem,
                          pretty_print=True).decode(
                              "utf-8"), sent_xml.decode("utf-8")



# -*- coding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""pass.

A CredentialsProvider which calls the external program 'pass' for the
password.
"""

from subprocess import check_output, CalledProcessError
from ...exceptions import ConfigurationError
from .interface import CredentialsAuthenticator
from .external_credentials_provider import ExternalCredentialsProvider


def get_authentication_provider():
    """get_authentication_provider.

    Return an authenticator which uses plain text username/password credentials.
    The difference to the `plain` module is that this implementation retrieves
    the password from the external program 'pass'.

    Returns
    -------
    CredentialsAuthenticator
        with a 'PassCaller' as back-end.
    """
    return CredentialsAuthenticator(PassCaller(callback=_call_pass))


def _call_pass(**config):
    if "password_identifier" not in config:
        raise ConfigurationError("Your configuration did not provide a "
                                 "`password_identifier` which is needed "
                                 "by the `PassCaller` to retrieve the "
                                 "password in question.")

    try:
        return check_output(
            "pass " + config["password_identifier"],
            shell=True).splitlines()[0].decode("UTF-8")
    except CalledProcessError as exc:
        raise RuntimeError(
            "Password manager returned error code {}. This usually "
            "occurs if the password_identifier in .pylinkahead.ini is "
            "incorrect or missing.".format(exc.returncode))


class PassCaller(ExternalCredentialsProvider):
    """PassCaller.

    A class for retrieving the password from the external program 'pass' and
    storing the username/password credentials as plain text strings.

    Methods
    -------
    configure

    Attributes
    ----------
    password
    username
    """
    # all the work is done in _call_pass and the super class
    pass



# -*- encoding: utf-8 -*-
#
# ** header v3.0
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
"""This module defines the CaosDBServerConnection interface."""
from abc import ABCMeta, abstractmethod, abstractproperty
from warnings import warn

# meta class compatible with Python 2 *and* 3:
ABC = ABCMeta('ABC', (object, ), {'__slots__': ()})


class CaosDBHTTPResponse(ABC):
    """An abstract class which defines a minimal interface for responses of the
    LinkAheadServer."""

    @abstractmethod
    def read(self, size=-1):
        """Read up to *size* bytes from the response body.

        If size is unspecified or -1, all bytes until EOF are returned.
        """

    @abstractmethod
    def getheader(self, name, default=None):
        """Return the value of the header *name* or the value of *default* if
        there is no such header.

        If there are multiple headers with that name, return all of the
        values joined by ', '. If default is an iterable, its elements
        are returned likewise.
        """

    @abstractproperty
    def status(self):
        """Status code of the response."""

    @abstractmethod
    def getheaders(self):
        """Return all headers."""

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        self.close()

    @abstractmethod
    def close(self):
        """close.

        Close this response. Depending on the implementation this might
        also close underlying streams, sockets etc.
        """


class CaosDBServerConnection(ABC):
    """Abstract class which defines the interface for sending requests to the
    LinkAhead server."""

    @abstractmethod
    def request(self, method, path, headers=None, body=None, **kwargs):
        """Abstract method. Implement this method for HTTP requests to the
        LinkAhead server.

        Returns
        -------
        CaosDBHTTPResponse
        """

    @abstractmethod
    def configure(self, **kwargs):
        """Configure the connection. This method is to be called by
        configure_connection.

        Returns
        -------
        None
        """
